#include <assert.h>
#include <inttypes.h>
#include <mpi.h>
#include <glib.h>
#include "coord-paic.h"
#include "common-paic.h"
#include "paic-collab.h"

enum MsgType {
    M_RENDERED,          /* "I'm done rendering." */
    M_REQUEST_COMPOSITE, /* "Will you composite with me?" */
    M_ACCEPT_COMPOSITE,  /* "Yes, I will accept your composite." */
    M_DENY_COMPOSITE     /* "No, I won't composite with you right now." */
};
enum State {
    S_WAIT_RENDER,       /* waiting for other nodes to finish rendering */
    S_WAIT_BUDDY,        /* sent M_REQUEST_COMPOSITE, waiting on response */
    S_WAIT_COOP,         /* waiting for cooperating thread to composite with an
                            accepted buddy. */
};
struct message {
    enum MsgType type;
    int source;
};

static MPI_Request *msgs;
/* Many times we only need to exchange metadata, which is contained entirely in
 * the message's source proc # and MPI tag.  Unfortunately MPI requires us to
 * always send data, and if we're sending data asynchronously that data needs
 * to stay around until the Irecv occurs.
 * So we just have a single `garbage' location which is always around, and send
 * it back and forth. */
static int *garbage;
static enum State cur_state;

static gboolean done();
static void send_completion_messages();
static struct message recv_update();
static void request_composite(int);
static void accept_composite(int);
static void deny_composite(int);
static void enqueue_buddy(int);

/** notify all processes that rendering has completed.
 * also serves as a sort of `initialization' for the composition. */
void coord_render_complete()
{
    garbage = malloc(sizeof(int));
    send_completion_messages();
    cur_state = S_WAIT_RENDER;
}

/** composition coordination.
 * Waits for messages from other nodes.  Chooses a node to composite with.
 * @returns the MPI id of the node to composite with */
int
coord_master()
{
    do {
        struct message msg = recv_update();
        switch(msg.type) {
            case M_RENDERED:
                SET_READY(msg.source);
                /* If we've got no ideas for a buddy yet, ask this guy. */
                if(cur_state == S_WAIT_RENDER) { /* a potential buddy! */
                    request_composite(msg.source);
                }
                break;
            case M_REQUEST_COMPOSITE:
                SET_READY(msg.source); /* might have seen b4 M_RENDERED. */
                if(cur_state == S_WAIT_RENDER) { /* a buddy! */
                    accept_composite(msg.source);
                    return msg.source;
                } else {
                    deny_composite(msg.source);
                }
                break;
            case M_ACCEPT_COMPOSITE:
                /* They shouldn't send an accept unless we've sent a request
                 * ... and we'll always be in a wait state if we sent that
                 * request. */
                assert(cur_state == S_WAIT_BUDDY);
                cur_state = S_WAIT_COOP;
                return msg.source; /* found a buddy -- bail! */
                break;
            case M_DENY_COMPOSITE:
                /* crap, need to choose someone new */
#if 0
                choose_someone_new();
#else
                g_error("fixme!");
#endif
                break;
        }
    } while(!done());
    return COORD_COMPLETE;
}

static gboolean
done()
{
    for(int i=0; i < pk_procs(); ++i) {
        if(!IS_READY(i)) {
            return FALSE;
        }
    }
    return TRUE;
}

static void
send_completion_messages()
{
    msgs = malloc(sizeof(MPI_Request) * pk_procs());
    for(int i=0; i < pk_procs(); ++i) {
        MPI_Isend(garbage, 1, MPI_INT, i, M_RENDERED, MPI_COMM_WORLD,
                  &(msgs[i]));
    }
    cur_state = S_WAIT_RENDER;
}

static struct message
recv_update()
{
    int idx;
    int recv;
    struct message msg;
    MPI_Status m_stat;

    MPI_Waitany(pk_procs(), msgs, &idx, &m_stat);
    MPI_Irecv(&recv, 1, MPI_INT, idx, m_stat.MPI_TAG, MPI_COMM_WORLD,
              &msgs[idx]);
    SET_READY(idx);

    msg.source = idx;
    msg.type = m_stat.MPI_TAG;
}

/** send an asynch message to this rank, asking if they'll composite with us.
 */
static void
request_composite(int from)
{
    MPI_Isend(garbage, 1, MPI_INT, from, M_REQUEST_COMPOSITE, MPI_COMM_WORLD,
              &msgs[from]);
    cur_state = S_WAIT_BUDDY;
}

/** send asynch message saying, `yes, I'll composite with you'. */
static void
accept_composite(int with)
{
    MPI_Isend(garbage, 1, MPI_INT, with, M_ACCEPT_COMPOSITE, MPI_COMM_WORLD,
              &msgs[with]);
    cur_state = S_WAIT_COOP;
}

/** send an asynch message to this rank, saying we can't composite with them
 * right now; maybe they should try again later. */
static void deny_composite(int to)
{
    MPI_Isend(garbage, 1, MPI_INT, to, M_DENY_COMPOSITE, MPI_COMM_WORLD,
              &msgs[to]);
}
