/**
 * \file hc.c
 * \brief HotCoffee management.
 * \author Tom Fogal
 * \date 2008-07-05
 */
#ifndef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 200112L
#endif /* !_POSIX_C_SOURCE */

#include <assert.h>
#include <err.h>
#include <stdio.h>
#include <GL/ice-t.h>
#include <GL/ice-t_mpi.h>

#include "hc.h"

/** This should really be external to HC ... */
#ifdef NDEBUG
#   define DEBUG_ONLY(x) /* nothing. */
#else
#   define DEBUG_ONLY(x) x
#endif

#define ICET_CHECK_ERROR                                                     \
    do {                                                                     \
        GLenum _err = icetGetError();                                        \
        const char *_es = NULL;                                              \
        switch(_err) {                                                       \
            case ICET_NO_ERROR: _es = "no error"; break;                     \
            case ICET_SANITY_CHECK_FAIL: _es = "sanity check failed"; break; \
            case ICET_INVALID_ENUM: _es = "invalid enum"; break;             \
            case ICET_BAD_CAST: _es = "bad cast"; break;                     \
            case ICET_OUT_OF_MEMORY: _es = "out of memory"; break;           \
            case ICET_INVALID_OPERATION: _es = "invalid operation"; break;   \
            case ICET_INVALID_VALUE: _es = "invalid value"; break;           \
        }                                                                    \
        if(_err != ICET_NO_ERROR) {                                          \
            warnx("[ICET_ERROR] %s\n", _es);                                 \
        }                                                                    \
    } while(0)

#define PR_ICET_MPI                                                         \
    do {                                                                    \
        GLint rank, nproc, ntiles;                                          \
        ICET(icetGetIntegerv(ICET_RANK, &rank));                            \
        ICET(icetGetIntegerv(ICET_NUM_PROCESSES, &nproc));                  \
        ICET(icetGetIntegerv(ICET_NUM_TILES, &ntiles));                     \
        warnx("IceT %8s %8s %8s\n   %8d %8d %8d\n",                         \
              "rank", "procs", "tiles", rank, nproc, ntiles);               \
    } while(0)

#ifdef NDEBUG
#   define ICET(x) x
#else
#   define ICET(x)                    \
    do {                              \
        x;                            \
        DEBUG_ONLY(ICET_CHECK_ERROR); \
    } while(0)
#endif

static IceTCommunicator comm;
static IceTContext context;

static void tile_setup(size_t w, size_t h);
static void write_icet_buf(const char *fn, size_t width, size_t height);
static void write_ppm(const char *filename, const GLubyte *buffer,
                      size_t width, size_t height);

void
hc_init()
{
    comm = icetCreateMPICommunicator(MPI_COMM_WORLD);
    ICET_CHECK_ERROR;
    context = icetCreateContext(comm);
    ICET_CHECK_ERROR;

    ICET(icetSetContext(context));

    ICET(icetDiagnostics(ICET_DIAG_FULL));

    ICET(icetStrategy(ICET_STRATEGY_PAIC));
    ICET(icetEnable(ICET_DISPLAY));

    ICET(icetInputOutputBuffers(
            ICET_COLOR_BUFFER_BIT | ICET_DEPTH_BUFFER_BIT, /* inputs */
            ICET_COLOR_BUFFER_BIT | ICET_DEPTH_BUFFER_BIT  /* outputs */
        ));

    DEBUG_ONLY(PR_ICET_MPI);
}

#include "ogl-surf.h"
void
hc_render(const struct gl_draw_fqn * const fqn)
{
    GLint rank;
    char ppm[1024];

    tile_setup(fqn->width, fqn->height);

    ICET(icetDrawFunc(fqn->gl_draw));
    ICET(icetDrawFrame());
    MPI_Barrier(MPI_COMM_WORLD);

    ICET(icetGetIntegerv(ICET_RANK, &rank));
    snprintf(ppm, 1024, "it-%d.ppm", rank);

    write_icet_buf(ppm, fqn->width, fqn->height);
    if(rank == 0) {
        void *buf = hack();
        write_ppm("hack0.ppm", buf, fqn->width, fqn->height);
    }
}

void
hc_deinit()
{
    ICET(icetDestroyContext(context));
    ICET(icetDestroyMPICommunicator(comm));
}

static void
tile_setup(size_t w, size_t h)
{
    ICET(icetResetTiles());
    ICET(icetAddTile(0,0, w,h, 0));
}

static void
write_icet_buf(const char *fn, size_t width, size_t height)
{
    GLint color_format;
    GLubyte *pixels = NULL;
    GLint available;
    GLint rank;

    icetGetIntegerv(ICET_COLOR_FORMAT, &color_format);
    if(color_format != GL_RGBA) {
        errx(EXIT_FAILURE, "color format is (confusingly) not RGBA...");
    }
    icetGetIntegerv(ICET_COLOR_BUFFER_VALID, &available);
    icetGetIntegerv(ICET_RANK, &rank);
    if(!available) {
        warnx("Color buffer not available on %d", rank);
        return;
    } else {
        printf("Reading color buffer on %d ...\n", rank);
    }

    pixels = icetGetColorBuffer();
    ICET_CHECK_ERROR;
    assert(pixels);

    write_ppm(fn, pixels, width, height);
}

/* taken from a Mesa demo */
void
write_ppm(const char *filename, const GLubyte *buffer,
          size_t width, size_t height)
{
    const int binary = 0;
    FILE *f = fopen(filename, "w");
    if (f) {
        int i, x, y;
        const GLubyte *ptr = buffer;
        if (binary) {
            fprintf(f,"P6\n");
            fprintf(f,"# ppm-file created by ppm.c\n");
            fprintf(f,"%u %u\n", width,height);
            fprintf(f,"255\n");
            fclose(f);
            f = fopen(filename, "ab");  /* reopen in binary append mode */
            for (y=height-1; y>=0; y--) {
                for (x=0; x<width; x++) {
                    i = (y*width + x) * 4;
                    fputc(ptr[i], f);   /* write red */
                    fputc(ptr[i+1], f); /* write green */
                    fputc(ptr[i+2], f); /* write blue */
                }
            }
        } else {
            /*ASCII*/
            int counter = 0;
            fprintf(f,"P3\n");
            fprintf(f,"# ascii ppm file created by hc.c\n");
            fprintf(f,"%u %u\n", width, height);
            fprintf(f,"255\n");
            for (y=height-1; y>=0; y--) {
                for (x=0; x<width; x++) {
                    i = (y*width + x) * 4;
                    fprintf(f, " %3d %3d %3d", ptr[i], ptr[i+1], ptr[i+2]);
                    counter++;
                    if (counter % 5 == 0)
                        fprintf(f, "\n");
                }
            }
        }
        fclose(f);
    }
}
