#include "util.h"

#include "posix.h"
#include <err.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <unistd.h>
#include <pthread.h>


static bool verbosity = false;
static bool tracing = false;
static FILE *trace_file = NULL;

/** malloc w/ fatal error if it fails.  I think I originally got this from some
 * GNU manual... */
G_GNUC_MALLOC void *
xmalloc(size_t size)
{
	void *block = malloc(size);
	assert(size > 0);
	trace("Xmalloc(%u) == %p", size, block);
	if(block == NULL && size != 0) {
		err(EXIT_FAILURE, "Out of memory.");
	}
	return block;
}

/** same deal, calloc though */
G_GNUC_MALLOC void *
xcalloc(size_t size)
{
	void *block = calloc(1, size);
	assert(size > 0);
	trace("xcalloc(%u) == %p", size, block);
	if(block == NULL && size != 0) {
		err(EXIT_FAILURE, "Out of memory.");
	}
	return block;
}

/** deallocates a block and sets it to NULL, to make sure we get an error if we
 * try to use it again. */
void
xfree(void *block)
{
	assert(block != NULL);
	free(block);
	block = NULL;
}

/** turns on verbosity.  default off. */
void verbose_on() { verbosity = true; }

/** turns off verbosity.  default is off. */
void verbose_off() { verbosity = false; }

/** gives the current verbosity value */
bool is_verbose() { return verbosity; }

/** prints out a message if verbosity is turned on */
void verbose(const char *fmt, ...)
{
	char buf[512];

	if(verbosity == true) {
		va_list ap;
		va_start(ap, fmt);
			vsprintf(buf, fmt, ap);
		va_end(ap);
		printf("%s\n", buf);
	}
}

/** makes trace() calls have an effect */
void
trace_enable(char *fn)
{
	tracing = true;
	trace_file = NULL;
	if(fn) {
		trace_file = fopen(fn, "w+");
	}
}

/** prints out a message if tracing */
void
trace(const char *fmt, ...)
{
	char buf[512];

	if(tracing == true) {
		va_list ap;
		va_start(ap, fmt);
			vsprintf(buf, fmt, ap);
		va_end(ap);
		chomp(buf);
		if(trace_file != NULL) {
			fprintf(trace_file, "{trace(%lu)} %s\n",
			        (unsigned long)pthread_self(), buf);
		}
		fprintf(stderr, "{trace(%lu)} %s\n", (unsigned long)pthread_self(), buf);
	}
}

void
trace_disable()
{
	fflush(stderr);
	if(trace_file) {
		fclose(trace_file);
	}
	tracing = false;
}

/** kind of like trace, except the server trace messages should always be
 * active -- they are not simply for debugging. */
void
sv_log(int fd, const char *fmt, ...)
{
	char buf[512];

	va_list ap;
	va_start(ap, fmt);
		vsprintf(buf, fmt, ap);
	va_end(ap);
#ifndef RB_CLIENT
#  if !defined(NDEBUG)
	  fprintf(stdout, "{log(%lu)} fd %-2d %s\n", (unsigned long)pthread_self(),
	                                           fd, buf);
#  else
	  fprintf(stdout, "fd %-2d %s\n", fd, buf);
#  endif
#endif
	fflush(stdout);
}

/** likewise, but now it IS for debugging. */
void
sv_log_verbose(int fd, const char *fmt, ...)
{
	char buf[512];

	va_list ap;
	va_start(ap, fmt);
		vsprintf(buf, fmt, ap);
	va_end(ap);
#ifndef NDEBUG
	fprintf(stdout, "{log(%lu)} fd %-2d %s\n", (unsigned long)pthread_self(),
	                                           fd, buf);
	fflush(stdout);
#endif
}

/** removes the newline from the end of a string, if there is one. */
char *chomp(char *s)
{
	if(s[strlen(s)-1] == '\n') {
		s[strlen(s)-1] = '\0';
	}
	return s;
}

/** the reverse of chomp -- ADDS a newline if one isn't there.  the string must
 * be big enough to handle one extra character. */
char *anti_chomp(char *s)
{
	size_t len = strlen(s);
	if(s[len-1] != '\n') {
		s[len] = '\n';
		s[len+1] = '\0';
	}
	return s;
}

/** searches for a newline in the given string and makes sure the character
 * preceding it is '/'.  If not, it adds that character, increasing the length 
 * of the string by 1. */
void
ensure_trailing_slash(char *s)
{
	size_t len = strlen(s);

	assert(s[len] == '\0');
	assert(s[len-1] == '\n');
	if(s[len-2] != '/') {
		s[len+1] = '\0';
		s[len] = '\n';
		s[len-1] = '/';
	}
}

/** gives the index of a particular string in an array of strings.  -1 if the
 * string is not found.
 * @param s the string to look for
 * @param arr the array to search in
 * @param sz size of the array
 * @return the index, or -1 if not found. */
int
find_string_idx(const char * const s, const char * const * const arr,
                unsigned int sz)
{
	unsigned int idx;

	for(idx=0; idx < sz; ++idx) {
		if(strcmp(s, arr[idx]) == 0) {
			return idx;
		}
	}
	return -1;
}

/** returns true if argument 1 is a prefix for argument 2 */
bool
prefix(const char *prefix, const char *s)
{
	size_t len = strlen(prefix);
	size_t i;

	if(len > strlen(s)) { return false; }

	for(i=0; i < len; ++i) {
		if(prefix[i] != s[i]) {
			return false;
		}
	}
	return true;
}

/** converts a string to an unsigned integer, checking for errors while doing
 * so.  errs out if the conversion fails */
uint64_t
str_to_uint(const char *s)
{
	unsigned long ret;
	const char *string;
	char *end;

	assert(strlen(s) >= 1);
	/* apparently strtoul() will happily convert a negative string into an
	 * unsigned integer.  Do a (simple) check for that and bail out if the
	 * number we're given is negative. */
	if(s[0] == '-') {
		errx(EXIT_FAILURE, "Illegal unsigned integer constant %s", s);
	}
	string = s;
	if(string[0] == '0' && string[1] != 'x') {
		/* get rid of leading 0's, they confuse strtoul */
		while(*string == '0') { string++; }
	}

	errno = 0;
	ret = strtoul(string, &end, 0);
	switch(errno) {
		case 0: /* success */ break;
		case ERANGE:
			{
				unsigned long int bytes = sizeof(unsigned long);
				err(EXIT_FAILURE, "Constant does not fit in %lu bytes", bytes);
			}
			break;
		case EINVAL:
			err(EXIT_FAILURE, "Conversion specifier unsupported");
			break;
		default:
			err(EXIT_FAILURE, "Cannot convert integer from string: %s",
			    strerror(errno));
			break;
	}
	if(end == s) {
		errx(EXIT_FAILURE, "Illegal unsigned integer constant %s", s);
	}
	if(ret < 0) {
		errx(EXIT_FAILURE, "Cannot convert %s to unsigned integer", s);
	}
	return (unsigned int) ret;
}

/** converts a string to an integer, checking for errors while doing
 * so.  errs out if the conversion fails */
int str_to_int(const char *s)
{
	long int ret;
	const char *string;
	char *end;

	string = s;
	if(string[0] == '0' && string[1] != 'x') {
		/* get rid of leading 0's, they confuse strtol */
		while(*string == '0') { string++; }
	}

	errno = 0;
	ret = strtol(string, &end, 0);
	switch(errno) {
		case 0: /* success */ break;
		case ERANGE:
			{
				unsigned long int bytes = sizeof(unsigned long);
				err(EXIT_FAILURE, "Constant does not fit in %lu bytes", bytes);
			}
			break;
		case EINVAL:
			err(EXIT_FAILURE, "Conversion specifier unsupported");
			break;
		default:
			err(EXIT_FAILURE, "Cannot convert integer from string: %s",
			    strerror(errno));
			break;
	}
	if(end == s) {
		err(EXIT_FAILURE, "Illegal integer constant %s", s);
	}
	if(ret < 0) {
		err(EXIT_FAILURE, "Cannot convert %s to integer", s);
	}
	return (int) ret;
}

G_GNUC_MALLOC char *
my_strdup(const char *s)
{
	char *ret = xmalloc(strlen(s) + 1);
	strcpy(ret, s);
	return ret;
}
