#ifndef TJF_UTIL_H
#define TJF_UTIL_H

#include "posix.h"
/** we have to enable getline() by saying we know what library we're using,
 * since it is a non-standard GNU extension. */
#if HAVE_GETLINE
#  define _GNU_SOURCE
#endif
#include <assert.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include <glib.h>

#include "types.h"

#define XMALLOC(sz)                                                   \
      xmalloc(sz);                                                    \
      trace("xmalloc(%d) in %s from %s", sz, __FUNCTION__, __FILE__);

#define XCALLOC(sz)                                                   \
      xcalloc(sz);                                                    \
      trace("xcalloc(%d) in %s from %s", sz, __FUNCTION__, __FILE__);

/** malloc w/ fatal error if it fails.  I think I originally got this from some
 * GNU manual... */
G_GNUC_MALLOC void *xmalloc(size_t size);

/** same deal, calloc though */
G_GNUC_MALLOC void *xcalloc(size_t size);

/** frees and sets the memory to NULL to help detect bugs */
void xfree(void *block);

/** xfree() with a helpful trace message. */
#define XFREE(x)                                                    \
   {                                                                \
       trace("xfree(%p) in %s from %s", x, __FUNCTION__, __FILE__); \
       xfree(x);                                                    \
   }

/** turns on verbosity.  default off. */
void verbose_on();
/** turns off verbosity.  default is off. */
void verbose_off();
/** gives the current verbosity value */
bool is_verbose();
/** prints out a message if verbosity is turned on */
void verbose(const char *fmt, ...);

/** similar functionality to verbose(), except only for tracing program
 * execution. */
/**@{*/
void trace_enable(char *fn);
void trace(const char *fmt, ...);
void trace_disable();
/**@}*/

/** kind of like trace, except the server trace messages should always be
 * active -- they are not simply for debugging. */
void sv_log(int fd, const char *fmt, ...);
/** likewise, but now it IS for debugging. */
void sv_log_verbose(int fd, const char *fmt, ...);

/** removes the newline from the end of a string, if there is one. */
char *chomp(char *);

/** the reverse of chomp -- ADDS a newline if one isn't there.  the string must
 * be big enough to handle one extra character. */
char *anti_chomp(char *);

/** searches for a newline in the given string and makes sure the character
 * preceding it is '/'.  If not, it adds that character, increasing the length 
 * of the string by 1. */
void ensure_trailing_slash(char *);

/** gives the index of a particular string in an array of strings.  -1 if the
 * string is not found. */
int find_string_idx(const char * const s, const char * const * const,
                    unsigned int sz);

/** returns true if argument 1 is a prefix for argument 2 */
bool prefix(const char *prefix, const char *s);

/** converts a string to an unsigned integer, checking for errors while doing
 * so.  errs out if the conversion fails */
uint64_t str_to_uint(const char *s);

/** converts a string to an integer, checking for errors while doing
 * so.  FATALs out if the conversion fails */
int str_to_int(const char *s);

/** apparently strdup made it into POSIX 2004, but not 2001? */
G_GNUC_MALLOC char *my_strdup(const char *s);
/** @todo some type of autoconf-style test to see if strdup() exists, use
 * system one if so. */
#define STRDUP my_strdup

/** portable macro to grab a line.  Attempts to use getline() if it can
 * (getline is quite fast), and drops back to fgets if getline isn't available.
 * fgets should always be available -- it's in POSIX and ANSI C -- but an
 * fscanf could work here too (albeit slowly).
 * @param fp the file pointer to read from
 * @param buf a char* buffer that must be dynamically allocated memory.  Will
 *            be mutated to contain the line.
 * @param length the length of `buf'.  Will be modified if `buf' is resized. */
#if HAVE_GETLINE
#  define GETLINE(fp, buf, length) { getline(&buf, &length, fp); }
#else
#  define GETLINE(fp, buf, length) { fgets(buf, length, fp); }
#endif

#endif /* TJF_UTIL_H */
